import oss2
from flask import Flask, g
from qcloud_cos import CosConfig, CosS3Client

COS_PARAMS_MAPPING = {
    "appid": "Appid",
    "region": "Region",
    "secret_id": "SecretId",
    "secret_key": "SecretKey",
    "token": "Token",
    "scheme": "Scheme",
    "timeout": "Timeout",
    "access_id": "SecretId",
    "access_key": "SecretKey",
    "endpoint": "Endpoint",
    "ip": "IP",
    "port": "Port",
    "anonymous": "Anonymous",
    "ua": "UA",
    "domain": "Domain",
    "service_domain": "ServiceDomain",
    "pool_connections": "PoolConnections",
    "pool_max_size": "PoolMaxSize",
    "allow_redirects": "AllowRedirects",
    "sign_host": "SignHost",
    "endpoint_ci": "EndpointCi",
}


def storage_cos_client() -> CosS3Client:
    return g.cos_client


def storage_oss_service() -> oss2.Service:
    return g.oss_service


def storage_oss_bucket() -> oss2.Bucket:
    return g.oss_bucket


def setup_storage(app: Flask):
    args = app.config.get_namespace("STORAGE_")
    vendor = args.pop("vendor")
    if vendor == "cos":
        new_args = {}
        for key, val in args.items():
            new_args[COS_PARAMS_MAPPING[key]] = val
        cos_config = CosConfig(**new_args)
        cos_client = CosS3Client(cos_config)

        @app.before_request
        def do_before_request():
            g.cos_client = cos_client
    elif vendor == "oss":
        auth = oss2.Auth(args.pop("access_key_id"), args.pop("access_key_secret"))
        service = oss2.Service(auth, endpoint=args.get("endpoint"))
        bucket = None
        if "bucket" in args:
            bucket = oss2.Bucket(auth, args.get("endpoint"), args.get("bucket"))

        @app.before_request
        def do_before_request():
            g.oss_service = service
            g.oss_bucket = bucket
    else:
        raise Exception('only cos, oss is supported')



from flask import Flask, g
from opentelemetry.instrumentation.sqlalchemy import SQLAlchemyInstrumentor
from sqlalchemy import create_engine
from sqlalchemy.engine import Engine
from sqlalchemy.orm import sessionmaker, registry, Session

model_registry = registry()

ModelBase = model_registry.generate_base(name="ModelBase")


def database_engine() -> Engine:
    return g.database_engine


def database_session_maker() -> sessionmaker:
    return g.database_session_maker


def database_session() -> Session:
    return database_session_maker()()


def database_transaction() -> Session:
    return database_session_maker().begin()


def setup_database(app: Flask, engine_args: dict = None, session_args: dict = None):
    if engine_args is None:
        engine_args = {}
    if session_args is None:
        session_args = {}

    engine_extras = app.config.get_namespace("DATABASE_ENGINE_")
    url = engine_extras.pop("url")
    engine_extras.update(**engine_args)

    engine = create_engine(url, **engine_extras)
    SQLAlchemyInstrumentor().instrument(engine=engine)

    session_extras = app.config.get_namespace("DATABASE_SESSION_")
    session_extras.update(**session_args)

    session_maker = sessionmaker(engine, **session_extras)

    @app.before_request
    def do_before_request():
        g.database_engine = engine
        g.database_session_maker = session_maker



from .celery import modern_celery
from .config import setup_config
from .database import setup_database, database_session_maker, database_engine, model_registry, ModelBase, \
    database_session, database_transaction
from .flask import modern_flask, halt
from .observability import setup_observability, ignore_metrics
from .redis import setup_redis, redis_client, redis_cached, redis_synchronized
from .storage import setup_storage, storage_cos_client, storage_oss_service, storage_oss_bucket



from flask import Flask


class HaltException(Exception):

    def __init__(self, *args):
        super().__init__("modern-flask halted")
        self.args = args


def halt(*args):
    raise HaltException(*args)


def modern_flask(*args, **kwargs) -> Flask:
    app = Flask(*args, **kwargs)

    @app.errorhandler(404)
    def handle_404(_):
        return {"message": "not found"}, 404

    @app.errorhandler(HaltException)
    def handle_halt_exception(e: HaltException):
        if len(e.args) == 1:
            return e.args[0]
        return tuple(e.args)

    return app



from celery import Celery
from flask import Flask


def modern_celery(app: Flask, *args, **kwargs) -> Celery:
    cfg = app.config.get_namespace("CELERY_")
    cfg.update(**kwargs)
    return Celery(*args, **cfg)



import json
from functools import wraps

import redis
from flask import Flask, g
from opentelemetry.instrumentation.redis import RedisInstrumentor
from redis import Redis


def redis_client() -> redis.Redis:
    return g.redis_client


def redis_cached(ex: int = 10, *keys):
    """
    decorate a function with redis cache with JSON serialization
    :param ex: expiration in seconds
    :param keys: cache keys, will be composed with dot
    :return:
    """

    def wrapper_builder(f):
        @wraps(f)
        def wrapper(*args, **kwargs):
            cache_key = "cache." + ".".join([str(key) for key in keys])
            raw = redis_client().get(cache_key)
            if raw is None:
                ret = f(*args, **kwargs)
                redis_client().set(cache_key, json.dumps(ret), ex=ex)
                return ret
            return json.loads(raw)

        return wrapper

    return wrapper_builder


def redis_synchronized(*lock_args, **lock_kwargs):
    """
    decorate a function with redis_client.lock
    :param lock_args:
    :param lock_kwargs:
    :return:
    """

    def wrapper_builder(f):
        @wraps(f)
        def wrapper(*args, **kwargs):
            with redis_client().lock(*lock_args, **lock_kwargs):
                return f(*args, **kwargs)

        return wrapper

    return wrapper_builder


def setup_redis(app: Flask):
    RedisInstrumentor().instrument()

    args = app.config.get_namespace("REDIS_")
    client = Redis(**args)

    @app.before_request
    def do_before_request():
        g.redis_client = client



import pathlib

from setuptools import setup

HERE = pathlib.Path(__file__).parent
README = (HERE / "README.md").read_text()
REQUIRED = (HERE / "requirements.txt").read_text().splitlines()

setup(
    name="modern-flask",
    version="202204b3",
    description="modern flask framework",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://gitee.com/modernapps/modern-flask",
    author="guoyk93",
    author_email="hi@guoyk.net",
    license="MIT",
    classifiers=[
        "Environment :: Web Environment",
        "Operating System :: OS Independent",
        "Topic :: Internet :: WWW/HTTP",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Framework :: Flask"
    ],
    packages=["modern"],
    include_package_data=True,
    install_requires=REQUIRED,
)



import base64
import os
import tempfile
from typing import List

import kubernetes
import requests
from flask import Flask
from kubernetes.client import V1ConfigMap, V1Secret

KUBERNETES_NAMESPACE_FILENAME = "/var/run/secrets/kubernetes.io/serviceaccount/namespace"


def config_read_namespace() -> str:
    with open(KUBERNETES_NAMESPACE_FILENAME) as f:
        return f.read().strip()


def config_from_pycode(app: Flask, content: str):
    with tempfile.NamedTemporaryFile(mode='w', suffix="config.py") as f:
        f.write(content)
        f.flush()
        app.config.from_pyfile(f.name)


def setup_config(app: Flask):
    src = os.getenv("CONFIG_SOURCE", "config.py")
    if src.startswith("kubernetes://"):
        items: List[str] = src.removeprefix("kubernetes://").split("/")
        if len(items) == 3:
            items.insert(0, config_read_namespace())
        if len(items) != 4:
            raise Exception("invalid CONFIG_SOURCE: " + src)
        kubernetes.config.load_config()

        namespace, kind, name, key = items
        if kind.startswith("configmap"):
            config: V1ConfigMap = kubernetes.client.CoreV1Api().read_namespaced_config_map(name=name,
                                                                                           namespace=namespace)
            config_from_pycode(app, config.data[key])
        elif kind.startswith("secret"):
            secret: V1Secret = kubernetes.client.CoreV1Api().read_namespaced_secret(name=name, namespace=namespace)
            config_from_pycode(app, base64.b64decode(secret.data[key]).decode("utf-8"))
        else:
            raise Exception("invalid CONFIG_SOURCE: " + src)
    elif src.startswith("http://") or src.startswith("https://"):
        config_from_pycode(app, requests.get(src).text)
    else:
        app.config.from_pyfile(src)



import os
from functools import wraps
from timeit import default_timer
from typing import Optional, List, Tuple

from flask import Flask, Response, g, request
from opentelemetry import trace
from opentelemetry.exporter.jaeger.thrift import JaegerExporter
from opentelemetry.exporter.zipkin.json import ZipkinExporter
from opentelemetry.instrumentation.flask import FlaskInstrumentor
from opentelemetry.instrumentation.logging import LoggingInstrumentor
from opentelemetry.instrumentation.requests import RequestsInstrumentor
from opentelemetry.propagate import set_global_textmap
from opentelemetry.propagators.b3 import B3MultiFormat
from opentelemetry.sdk.resources import Resource
from opentelemetry.sdk.trace import TracerProvider
from opentelemetry.sdk.trace.export import BatchSpanProcessor
from opentelemetry.semconv.resource import ResourceAttributes
from prometheus_client import Histogram
from prometheus_client.exposition import make_wsgi_app


def ignore_metrics(f):
    @wraps(f)
    def wrapper(*args, **kwargs):
        g.metrics_ignore = True
        return f(*args, **kwargs)

    return wrapper


class Metrics:

    def __init__(self, app: Flask, metrics_path="/metrics", ignored_paths: Optional[List[str]] = None):
        self.metrics_path = metrics_path
        self.metrics_app = make_wsgi_app()
        self.ignored_paths = ignored_paths if ignored_paths is not None else []

        self.duration_histogram = Histogram(
            name="http_request_duration_seconds",
            documentation="duration of HTTP requests in seconds",
            labelnames=("method", "handler", "status"),
            buckets=Histogram.DEFAULT_BUCKETS,
        )

        @app.before_request
        def do_before_request():
            if self.should_ignore():
                return
            g.metrics_start_time = default_timer()

        @app.after_request
        def do_after_request(response: Response):
            if self.should_ignore():
                return response
            self.duration_histogram.labels(*self.calculate_request_labels(response)).observe(
                self.calculate_request_duration())
            return response

        @app.teardown_request
        def do_teardown_request(exception=None):
            if exception is None or self.should_ignore():
                return
            self.duration_histogram.labels(*self.calculate_request_labels(None)).observe(
                self.calculate_request_duration())

    def calculate_request_duration(self) -> float:
        _ = self
        return max(g.metrics_start_time - default_timer(), 0)

    def calculate_request_labels(self, response: Response = None) -> Tuple[str, str, str]:
        _ = self
        status = "500"
        if status is not None:
            status = str(response.status_code)
        rule = "/404"
        if request.url_rule is not None:
            rule = request.url_rule.rule
        return request.method, rule, status

    def should_ignore(self) -> bool:
        if hasattr(g, "metrics_ignore") and g.metrics_ignore:
            return True
        if any((request.url == p or (request.url_rule is not None and request.url_rule.rule == p)) for p in
               self.ignored_paths):
            g.metrics_ignore = True
            return True
        return False

    def expose(self, app: Flask):
        """
        exponse metrics to Flask application

        :param app: Flask application
        :return: None
        """

        @app.get(self.metrics_path)
        @ignore_metrics
        def route_metrics():
            return self.metrics_app


def setup_observability(app: Flask):
    # using B3 headers
    set_global_textmap(B3MultiFormat())

    # using jaeger exporter
    vendor = os.getenv("OTEL_EXPORTER", "zipkin")
    if vendor == "jaeger":
        span_exporter = JaegerExporter()
    elif vendor == "zipkin":
        span_exporter = ZipkinExporter()
    else:
        raise Exception("unsupported opentelemetry exporter: " + vendor)

    span_processor = BatchSpanProcessor(span_exporter)

    # setup global tracer provider
    service_name = os.getenv("SERVICE_NAME", "unknown-service")
    namespace = os.getenv("POD_NAMESPACE", "unknown-namespace")
    name = os.getenv("POD_NAME", "unknown-name")
    node_name = os.getenv("NODE_NAME", "unknown-node")
    tracer_provider = TracerProvider(resource=Resource.create({
        ResourceAttributes.SERVICE_NAME: service_name,
        ResourceAttributes.SERVICE_NAMESPACE: namespace,
        ResourceAttributes.SERVICE_INSTANCE_ID: name,
        ResourceAttributes.K8S_POD_NAME: name,
        ResourceAttributes.K8S_NAMESPACE_NAME: namespace,
        ResourceAttributes.K8S_NODE_NAME: node_name,
    }))
    tracer_provider.add_span_processor(span_processor)
    trace.set_tracer_provider(tracer_provider)

    # initialize tracing components
    LoggingInstrumentor().instrument(set_logging_format=True)
    FlaskInstrumentor().instrument_app(app)
    RequestsInstrumentor().instrument()

    # setup metrics
    Metrics(
        app,
        ignored_paths=[
            "/favicon.ico",
            "/healthz",
        ],
    ).expose(app)
